import pathlib
import sys
import venv

import pytest


@pytest.fixture
def executable():
    """Return the current interpreter's path."""
    return pathlib.Path(sys.executable)


@pytest.fixture(scope="session")
def full_venv(tmp_path_factory):
    """Create a virtual environment via venv."""
    symlinks = sys.platform != "win32"
    venv_path = tmp_path_factory.mktemp("venvs") / "full_venv"
    venv.create(
        venv_path, symlinks=symlinks, with_pip=False, system_site_packages=False
    )
    return venv_path



import os
import pathlib
import subprocess
import sys

import pytest

import microvenv
import microvenv._create

if sys.platform == "win32":
    pytest.skip("Windows is not supported.", allow_module_level=True)


@pytest.fixture
def base_executable():
    try:
        return pathlib.Path(sys._base_executable)
    except AttributeError:
        return pathlib.Path(sys.executable)


@pytest.fixture(scope="session")
def micro_venv(tmp_path_factory):
    venv_path = tmp_path_factory.mktemp("venvs") / "micro_venv"
    microvenv.create(venv_path)
    return venv_path


def test_code_size(executable, monkeypatch, tmp_path):
    """Make sure the source code can fit into `argv` for use with `-c`."""
    with open(microvenv._create.__file__, "r", encoding="utf-8") as file:
        source = file.read()
    monkeypatch.chdir(tmp_path)
    env_path = pathlib.Path(microvenv.DEFAULT_ENV_DIR)
    subprocess.check_call([os.fsdecode(executable), "-c", source])

    # Since `__name__ == "__main__"` calls `_create.main()`, don't worry about
    # validating the virtual environment details as the CLI tests take care of
    # that.
    assert env_path.is_dir()
    command = microvenv.parse_config(env_path)["command"]
    assert command.startswith(sys.executable)
    assert " -c " in command


def test_structure(full_venv, micro_venv):
    for dirpath, dirnames, filenames in os.walk(full_venv):
        root = pathlib.Path(dirpath)
        relative_root = root.relative_to(full_venv)
        micro_root = micro_venv / relative_root
        for dirname in dirnames:
            micro_dir = micro_root / dirname
            assert micro_dir.is_dir()

        for filename in filenames:
            # Don't care about activation scripts.
            if filename.lower().startswith("activate"):
                continue
            micro_file = micro_root / filename
            full_file = root / filename
            if full_file.is_file():
                assert micro_file.is_file()
            # Make sure that e.g. `python` is properly symlinked.
            elif full_file.is_symlink():
                assert micro_file.is_symlink()
                assert micro_file.resolve() == full_file.resolve()


def test_lib64(full_venv, micro_venv):
    micro_lib64 = micro_venv / "lib64"
    if not (full_venv / "lib64").exists():
        assert not micro_lib64.exists()
    else:
        assert micro_lib64.exists()
        assert micro_lib64.is_symlink()
        assert micro_lib64.resolve() == (micro_venv / "lib")
        # pathlib.Path.readlink() was added in Python 3.9.
        assert os.readlink(micro_lib64) == "lib"


@pytest.mark.parametrize(
    "key",
    ["include-system-site-packages", "version"],
)
def test_pyvenvcfg_data(full_venv, micro_venv, key):
    full_config = microvenv.parse_config(full_venv)
    micro_config = microvenv.parse_config(micro_venv)

    # Use the full config as source of keys to check as as we may have keys in the micro
    # venv that too new for the version of Python being tested against.
    if key in full_config:
        assert key in micro_config
        assert full_config[key] == micro_config[key]


def test_pyvenvcfg_home(base_executable, full_venv, micro_venv):
    full_config = microvenv.parse_config(full_venv)
    micro_config = microvenv.parse_config(micro_venv)

    assert full_config["home"] == os.fsdecode(base_executable.parent)  # Sanity check.
    assert micro_config["home"] == os.fsdecode(base_executable.parent)


def test_pyvenvcfg_executable(base_executable, full_venv, micro_venv):
    resolved_base_executable = base_executable.resolve()
    executable_path = os.fsdecode(resolved_base_executable)
    full_config = microvenv.parse_config(full_venv)
    if "executable" not in full_config:
        # Introduced in Python 3.11.
        pytest.skip("`executable` key not in pyvenv.cfg")

    micro_config = microvenv.parse_config(micro_venv)

    assert full_config["executable"] == executable_path  # Sanity check.
    assert micro_config["executable"] == executable_path


def test_pyvenvcfg_command(executable, micro_venv):
    config = microvenv.parse_config(micro_venv)
    script_path = pathlib.Path(microvenv._create.__file__).resolve()
    assert config["command"] == f"{executable} {script_path} {micro_venv.resolve()}"


def test_pyvenvcfg_command_relative(executable, monkeypatch, tmp_path):
    monkeypatch.chdir(tmp_path)
    venv_path = tmp_path / "venv"
    microvenv.create(pathlib.Path(venv_path.name))
    script_path = pathlib.Path(microvenv._create.__file__).resolve()
    config = microvenv.parse_config(venv_path)
    assert config["command"] == f"{executable} {script_path} {venv_path.resolve()}"



from os import PathLike
from typing import Iterable

DEFAULT_ENV_DIR: str

def create(
    env_dir: str | PathLike[str] = ..., *, scm_ignore_files=Iterable[str]
) -> None: ...
def main() -> None: ...



from os import PathLike, _Environ
from typing import TypedDict

IN_VIRTUAL_ENV: bool

class ActivationError(Exception): ...

def parse_config(env_dir: str | PathLike[str]) -> dict[str, str]: ...

class _ActivationEnvVars(TypedDict):
    PATH: str
    VIRTUAL_ENV: str

def activation(env_vars: _Environ[str] = ...) -> _ActivationEnvVars: ...



import os
import subprocess
import sys

import pytest

import microvenv

if sys.platform == "win32":
    pytest.skip("Windows is not supported.", allow_module_level=True)


@pytest.fixture
def CLI(executable):
    def _CLI(*args):
        subprocess.check_call(
            [os.fsdecode(executable), microvenv._create.__file__, *args]
        )

    return _CLI


@pytest.mark.parametrize(
    ["args", "expected_dir"], [([], ".venv"), (["some-venv"], "some-venv")]
)
def test_relative_path(monkeypatch, tmp_path, args, expected_dir, CLI):
    """Test using a relative path (both the default and explicitly provided)."""
    path = tmp_path / expected_dir
    monkeypatch.chdir(tmp_path)
    CLI(*args)
    assert path.is_dir()
    assert (path / "pyvenv.cfg").is_file()


def test_absolute_path(tmp_path, CLI):
    path = tmp_path / "some-venv"
    CLI(os.fsdecode(path))

    assert path.is_dir()
    assert (path / "pyvenv.cfg").is_file()


def test_too_many_args(tmp_path, CLI):
    path = tmp_path / "some-venv"
    with pytest.raises(subprocess.CalledProcessError):
        CLI(os.fsdecode(path), "extra-arg")

    assert not path.exists()


def test_default_scm_ignore_files(tmp_path, CLI):
    venv_path = tmp_path / "some-venv"
    CLI(os.fsdecode(venv_path))
    gitignore_path = venv_path / ".gitignore"

    assert gitignore_path.is_file()
    assert gitignore_path.read_text(encoding="utf-8") == "*\n"


def test_without_scm_ignore_files(tmp_path, CLI):
    venv_path = tmp_path / "some-venv"
    CLI("--without-scm-ignore-files", os.fsdecode(venv_path))
    gitignore_path = venv_path / ".gitignore"

    assert not gitignore_path.exists()



import sys

from ._create import main

if __name__ == "__main__":
    if sys.platform == "win32":
        print("Microvenv's CLI is not supported on Windows", file=sys.stderr)
        sys.exit(1)

    main()



import os
import pathlib
import sys

import pytest

import microvenv


def test_no_venv(monkeypatch):
    monkeypatch.setattr(microvenv, "IN_VIRTUAL_ENV", False)
    with pytest.raises(microvenv.ActivationError):
        microvenv.activation()


def test_PYTHONHOME():
    with pytest.raises(microvenv.ActivationError):
        microvenv.activation({"PYTHONHOME": "/home/user"})


def test_no_PATH():
    bin_dir = os.fsdecode(pathlib.Path(sys.executable).parent)
    env_vars = microvenv.activation({})

    assert env_vars["PATH"] == bin_dir


def test_PATH_prepended():
    path_var = os.pathsep.join(["A", "B", "C"])
    bin_dir = os.fsdecode(pathlib.Path(sys.executable).parent)
    env_vars = microvenv.activation({"PATH": path_var})

    assert env_vars["PATH"] == os.pathsep.join([bin_dir, path_var])


def test_VIRTUAL_ENV():
    env_path = pathlib.Path(sys.executable).parent.parent
    env_vars = microvenv.activation({})

    assert env_vars["VIRTUAL_ENV"] == os.fsdecode(env_path)


def test_default_env_vars():
    bin_path = pathlib.Path(sys.executable).parent
    bin_dir = os.fsdecode(bin_path)
    path = os.pathsep.join([bin_dir, os.environ["PATH"]])
    virtual_env = os.fsdecode(bin_path.parent)
    env_vars = microvenv.activation()

    assert env_vars["PATH"] == path
    assert env_vars["VIRTUAL_ENV"] == virtual_env



import os
import pathlib
import sys

# Exported as part of the public API.
from ._create import DEFAULT_ENV_DIR as DEFAULT_ENV_DIR

if sys.platform != "win32":
    from ._create import create as create

# https://docs.python.org/3/library/venv.html#how-venvs-work
IN_VIRTUAL_ENV = sys.prefix != sys.base_prefix


class ActivationError(Exception):
    """An error trying to calculate environment activation via :func:`activation`."""


def parse_config(env_dir):
    """Parse the ``pyvenv.cfg`` file in the specified virtual environment.

    A dict is returned with strings for keys and values. All keys are
    lowercased, but otherwise no validation is performed. No changes are made to
    the values (e.g., ``include-system-site-packages`` is not converted to a
    boolean, nor lowercased).

    Parsing is done in a way identical to how the :py:mod:`site` module does it.
    This means that ANY line with an ``=`` sign is considered a line with a
    key/value pair on it. As such, all other lines are treated as if they are
    comments. But this also means that having a line start with e.g., ``#`` does
    not signify a comment either if there is a ``=`` in the line.
    """
    config = {}
    venv_path = pathlib.Path(env_dir)
    with open(venv_path / "pyvenv.cfg", "r", encoding="utf-8") as file:
        # This is how `site` parses `pyvenv.cfg`, so it's about as
        # official as we can get.
        for line in file:
            if "=" in line:
                key, _, value = line.partition("=")
                config[key.strip().lower()] = value.strip()
    return config


# There are three possible APIs for helping with virtual environment activation:
# 1. Mutate the environment variable dictionary in-place;
#    ``activate(os.environ.copy())``
# 2. Return a mutated copy; ``activated_env_copy()``
# 3. Calculate the environment variables to add; ``os.environ | activation()``
#
# Option 1 has the issue that if you forget to pass in a copy you will end up
# mutating the running process' environment (in most cases).
# Option 2 has the issue of needing an awkward name to communicate what exactly
# is occurring.
# Option 3 has the issue of not being able to delete environment variables
# (i.e., PYTHONHOME).
#
# The decision was to go with option 3 with an exception if PYTHONHOME is set as
# it's the most self-documenting for the common case of PYTHONHOME not already
# being set.
def activation(env_vars=os.environ):
    """Returns a dict with env vars to activate the current virtual environment.

    If the current interpreter is not from a virtual environment,
    :exc:`ActivationError` is raised.

    Since this API is designed to be additive to a environment variables dict
    (i.e., ``os.environ | activation()``), key deletions are not supported. As
    such, :exc:`ActivationError` is raised if ``PYTHONHOME`` is found to be set.

    No environment variables are provided in the returned dict in relation to
    shell prompts.
    """
    if not IN_VIRTUAL_ENV:
        raise ActivationError("Not running from a virtual environment")
    elif "PYTHONHOME" in env_vars:
        raise ActivationError("PYTHONHOME is set")

    bin_path = pathlib.Path(sys.executable).parent
    bin_dir = os.fsdecode(bin_path)
    if "PATH" not in env_vars:
        path = bin_dir
    else:
        path = os.pathsep.join([os.fsdecode(bin_dir), env_vars["PATH"]])

    return {"PATH": path, "VIRTUAL_ENV": os.fsdecode(bin_path.parent)}



import os
import pathlib

import nox  # type: ignore


@nox.session(python=["3.7", "3.8", "3.9", "3.10", "3.11", "3.12"])
def test(session):
    session.install(".[test]")
    session.run("pytest")


@nox.session
def lint(session):
    session.install(".[lint]")
    session.run("ruff", "check", ".")
    session.run("black", "--check", ".")
    session.run("mypy", ".")
    session.run("stubtest", "microvenv")


@nox.session
def docs(session):
    docs_path = pathlib.Path("docs")
    session.install(".[doc]")
    session.debug("Recording `microvenv --help`")
    out = session.run("python", "-m", "microvenv", "--help", silent=True)
    with open(docs_path / "help.txt", "w", encoding="utf-8") as file:
        file.write(out.strip())
    session.run(
        "sphinx-build",
        "-j",
        "auto",
        "-W",
        os.fsdecode(docs_path),
        os.fsdecode(docs_path / "_build"),
    )



"""A minimal, self-contained implementation of `venv`.

Because this module is self-contained from the rest of the package, you can
execute this module directly. It has its own CLI (which can be explored via
`--help`).
"""
import argparse
import os
import pathlib
import sys
import sysconfig

# This should not change during execution, so it's reasonable as a global.
# The path is purposefully unresolved for pyvenv.cfg purposes.
_BASE_EXECUTABLE = pathlib.Path(getattr(sys, "_base_executable", sys.executable))

_PYVENVCFG_TEMPLATE = f"""\
home = {_BASE_EXECUTABLE.parent}
include-system-site-packages = false
version = {'.'.join(map(str, sys.version_info[:3]))}
executable = {_BASE_EXECUTABLE.resolve()}
command = {{command}}
"""


def _sysconfig_path(name, env_dir):
    variables = {
        "base": env_dir,
        "platbase": env_dir,
        "installed_base": env_dir,
        "installed_platbase": env_dir,
    }

    return pathlib.Path(sysconfig.get_path(name, "venv", variables))


DEFAULT_ENV_DIR = ".venv"


# Analogous to `venv.create()`.
def create(env_dir=DEFAULT_ENV_DIR, *, scm_ignore_files=frozenset(["git"])):
    """Create a minimal virtual environment.

    Analogous to `venv.create(env_dir, symlinks=True, with_pip=False)`.
    """
    env_path = pathlib.Path(env_dir)
    # sysconfig scheme support introduced in Python 3.11.
    try:
        scripts_dir = _sysconfig_path("scripts", env_path)
        include_dir = _sysconfig_path("include", env_path)
        purelib_dir = _sysconfig_path("purelib", env_path)
    except KeyError:
        scripts_dir = env_path / "bin"
        include_dir = env_path / "include"
        purelib_dir = (
            env_path
            / "lib"
            / f"python{sys.version_info.major}.{sys.version_info.minor}"
            / "site-packages"
        )
    for dir in (scripts_dir, include_dir, purelib_dir):
        dir.mkdir(parents=True)

    if sys.maxsize > 2**32 and os.name == "posix" and sys.platform != "darwin":
        lib_path = env_path / "lib"
        lib64_path = env_path / "lib64"
        # There is no guarantee the sysconfig scheme will produce a `lib`
        # directory.
        if lib_path.is_dir() and not lib64_path.exists():
            lib64_path.symlink_to("lib", target_is_directory=True)

    for executable_name in (
        "python",
        f"python{sys.version_info.major}",
        f"python{sys.version_info.major}.{sys.version_info.minor}",
    ):
        (scripts_dir / executable_name).symlink_to(_BASE_EXECUTABLE)

    if __spec__ is None:
        command = f"{sys.executable} -c '...'"
    else:
        module_path = pathlib.Path(__spec__.origin).resolve()
        command = f"{sys.executable} {module_path} {env_path.resolve()}"
    (env_path / "pyvenv.cfg").write_text(
        _PYVENVCFG_TEMPLATE.format(command=command),
        encoding="utf-8",
    )

    scm_ignore_files = frozenset(scm_ignore_files)
    if scm_ignore_files == {"git"}:
        (env_path / ".gitignore").write_text("*\n", encoding="utf-8")
    elif scm_ignore_files:
        unexpected = scm_ignore_files - {"git"}
        raise NotImplementedError(
            f"Only `git` is supported as a SCM ignore file, not {unexpected}."
        )


def main():
    parser = argparse.ArgumentParser(
        description="Create a minimal virtual environment."
    )
    parser.add_argument(
        "--without-scm-ignore-files",
        dest="scm_ignore_files",
        action="store_const",
        const=frozenset(),
        default=frozenset(["git"]),
        help="Skips adding SCM ignore files to the environment "
        "directory (otherwise a `.gitignore` file is added).",
    )
    parser.add_argument(
        "env_dir",
        default=DEFAULT_ENV_DIR,
        nargs="?",
        help=(
            "Directory to create virtual environment in "
            f"(default: {DEFAULT_ENV_DIR!r}"
        ),
    )
    args = parser.parse_args()
    create(args.env_dir, scm_ignore_files=args.scm_ignore_files)


if __name__ == "__main__":
    main()



import contextlib

import pytest

import microvenv


@contextlib.contextmanager
def write_config(venv_path, data):
    """Context manager to write the pyvenv.cfg and then restore it."""
    config_path = venv_path / "pyvenv.cfg"
    original_config = config_path.read_text(encoding="utf-8")
    config_path.write_text(data, encoding="utf-8")
    try:
        yield config_path
    finally:
        config_path.write_text(original_config, encoding="utf-8")


@pytest.mark.parametrize("equals", ["=", " = ", "= ", " ="])
def test_formatting_around_equals(full_venv, equals):
    with write_config(full_venv, f"key{equals}value\n"):
        config = microvenv.parse_config(full_venv)

        assert config["key"] == "value"


def test_comments(full_venv):
    with write_config(full_venv, "# A comment\nkey = value\n"):
        config = microvenv.parse_config(full_venv)

        assert config["key"] == "value"


def test_multiple_equals(full_venv):
    with write_config(full_venv, "key = value=value\n"):
        config = microvenv.parse_config(full_venv)

        assert config["key"] == "value=value"


def test_lowercase_keys(full_venv):
    with write_config(full_venv, "Key = value\n"):
        config = microvenv.parse_config(full_venv)

        assert config["key"] == "value"
